SystemOrganization addCategory: #ClojureST!

TestCase subclass: #CjAgentTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'ClojureST'!

!CjAgentTest methodsFor: 'configuration' stamp: 'StefanMarr 1/5/2012 14:24'!
agentNew
	^ CjAgent new! !

!CjAgentTest methodsFor: 'configuration' stamp: 'StefanMarr 1/5/2012 14:25'!
agentNew: initialState
	^ CjAgent new: initialState! !

!CjAgentTest methodsFor: 'running' stamp: 'StefanMarr 7/31/2012 23:17'!
tearDown
	(Smalltalk classNamed: #ONOMirror)
		ifNotNilDo: [:c | c reenableScheduling].
	^ super tearDown! !

!CjAgentTest methodsFor: 'tests' stamp: 'StefanMarr 1/5/2012 14:27'!
testAwait
	| agent |
	agent := self agentNew.

	agent send: [:i | #noop].
	agent send: [:i | #noop].
	agent send: [:i | #noop].
	agent send: [:i | #last].
	
	agent await.
	
	self assert: #last equals: agent read.
	! !

!CjAgentTest methodsFor: 'tests' stamp: 'StefanMarr 1/5/2012 14:27'!
testRead
	| agent |
	
	"without initial value"
	agent := self agentNew.
	self assert: nil    equals: agent read.
	
	"with initial value"
	agent := self agentNew: #first.
	self assert: #first equals: agent read.
	
	agent send: [:i | #foo].
	agent await.
	self assert: #foo equals: agent read.! !

!CjAgentTest methodsFor: 'tests' stamp: 'StefanMarr 1/5/2012 14:27'!
testSend
	| agent |
	agent := self agentNew: 0.
	
	self assert: 0 equals: agent read.
	
	agent send: [:old | old + 1].
	
	agent await.
	
	self assert: 1 equals: agent read.
	
	agent send: [:old | old + 20].
	agent send: [:old | old + 21].
	
	agent await.
	
	self assert: 42 equals: agent read.
	! !

!CjAgentTest methodsFor: 'tests' stamp: 'StefanMarr 1/5/2012 14:32'!
testSendWith
	| agent incFn |
	agent := self agentNew: 0.
	incFn := [:old :inc | old + inc].
	
	self assert: 0 equals: agent read.
	
	agent send: incFn with: 1.
	
	agent await.
	
	self assert: 1 equals: agent read.
	
	agent send: incFn with: 20.
	agent send: incFn with: 21.
	
	agent await.
	
	self assert: 42 equals: agent read.
	! !

!CjAgentTest methodsFor: 'tests' stamp: 'StefanMarr 1/5/2012 14:27'!
testShutdown
	| agent |
	agent := self agentNew.
	
	self deny: (agent instVarNamed: #process) isTerminated.
	
	agent shutdown.
	
	agent await.
	
	self assert:  (agent instVarNamed: #process) isTerminated.! !

Object subclass: #CjAgent
	instanceVariableNames: 'process mailbox state terminate'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'ClojureST'!

!CjAgent class methodsFor: 'instance creation' stamp: 'StefanMarr 1/5/2012 13:42'!
new: initialState
	| agent |
	agent := super new.
	agent instVarNamed: #state put: initialState.
	^ agent! !

!CjAgent methodsFor: 'agents' stamp: 'StefanMarr 1/5/2012 14:02'!
await
	"Block until all actions have been processed (weaker semantics than Clojure's)"
	
	[mailbox isEmpty] 
		whileFalse: [
			(Delay forMilliseconds: 50) wait.
		]! !

!CjAgent methodsFor: 'initialization' stamp: 'StefanMarr 1/7/2012 12:38'!
initialize
	super initialize.
	
	mailbox := SharedQueue new.
	terminate := false.
	
	process := [
		[terminate and: [mailbox isEmpty]] whileFalse: [
			self processIncomingMessages.
		]] fork.
	process name: 'Agent'.! !

!CjAgent methodsFor: 'private' stamp: 'StefanMarr 1/6/2012 15:40'!
processIncomingMessages
	| updateBlock |
	updateBlock := mailbox peekWaiting. "Only wait for the element, but dont remove it."
	
	state := updateBlock value: state.

	mailbox next. "Only remove it now, otherwise we get races with await."! !

!CjAgent methodsFor: 'agents' stamp: 'StefanMarr 1/5/2012 14:07'!
read
	^ state! !

!CjAgent methodsFor: 'agents' stamp: 'StefanMarr 1/5/2012 14:10'!
send: anUpdateBlock
	mailbox nextPut: anUpdateBlock.! !

!CjAgent methodsFor: 'agents' stamp: 'StefanMarr 1/5/2012 14:30'!
send: anUpdateBlock with: args
	self send: [:old | anUpdateBlock value: old value: args ]! !

!CjAgent methodsFor: 'agents' stamp: 'StefanMarr 1/5/2012 13:52'!
shutdown
	"Make sure that the process loop gets around to see that terminate is set to true.
	 So, we just set it ourselves.
	 No race conditions here, purely sequential."
	mailbox nextPut: [:ignore | terminate := true].! !
